@@ -7,7 +7,8 @@ from .exercise_regexp import ExerciseRegexp, ExerciseRegexpGroups
 from .exercise_generator import ExerciseGenerator
 from .exercise_class import ExerciseClass, ClassScenario, ClassExpression, ClassStatement
 
-from .content import TextContent, MarkdownContent
-from .quiz import Quiz, QuizQuestion, Option, CodeOption, MathOption, MarkdownOption
+from .content import TextContent, MarkdownContent, CodeContent
+from .quiz import Quiz, QuizQuestion, Option, CodeOption, MathOption, MarkdownOption, Explanation
+from .quiz_loader import run_yaml_quiz
 
 from .version import __version__
