@@ -5,22 +5,43 @@ from enum import Enum
 
 from ipywidgets import Layout, HBox, VBox, Checkbox, Button, HTML, HTMLMath
 
-from .content import Content, TextContent, CssContent
-from .storage import log_quiz, storage_read, storage_save
+from .content import Content, TextContent, CssContent, MarkdownContent
+from .storage import log2_quiz, storage_read, storage_save
 from .helpers import truncate
 
 CSS = """
+:root {
+    /* these two are for unanswered */
+    --question-bg-odd: #d6e4f0; 
+    --question-bg-even: #ddebf8;
+    --question-bg-right: #dafcf0;
+    --question-bg-wrong: #ffd6d9;
+    /* individual options after revealed */
+    --option-wrong: #ff2e63;
+    --explanation-right: #f8f8f8;
+    --explanation-wrong: #f0f0f0;
+    /* misc */
+    --question-header-bg: white;
+    --submit-bg: #c2f0fc;
+    /* borders */
+    --border-question: 2px solid #084177;
+    --border-code: 1px solid #a0a0a0;
+    --border-explanation: 1px solid #b0b0b0;
+    --border-submit: 2px solid #084177;
+}
+
 .widget-vbox.nbae-question, .widget-hbox.nbae-question {
     padding: 10px;
     border-radius: 10px;
     margin-bottom: 5px;
 }
 .nbae-question .question {
-    border: 2px solid #084177;
-    border-radius: 4px;
+    border: var(--border-question);
+    border-radius: 6px;
     width: max-content;
     max-width: 100%;
     padding: 4px 8px;
+    background-color: var(--question-header-bg);
 }
 /* question in a QuizQuestion with a vertical layout */
 .nbae-question.widget-vbox>.question {
@@ -31,12 +52,15 @@ CSS = """
     font-weight: bold;
 }
 
-.nbae-question .question:not(.exactly-one) .header::after {
+.nbae-question .question:not(.exactly-one) .question-header::after {
     content: "♧";
     align-self: center;
     font-weight: bold;
     font-size: 125%;
 }
+.nbae-question .question-header {
+    border-bottom: 1px solid #084177;
+}
 .nbae-question .score::before {
     content: "|";
     align-self: center;
@@ -56,32 +80,41 @@ CSS = """
     background-color: #202020;
 }
 .nbae-question .code {
-    border: 1px solid #aaa;
+    border: var(--border-code);
 }
 
-.nbae-question .widget-checkbox, .nbae-question .question2 {
+.nbae-question .widget-checkbox {
     width: auto;
     padding-left: 10px;
 }
 
+.nbae-question .question-sequel {
+    width: auto;
+    margin-left: 10px;
+    padding: 8px 10px;
+    /*border: var(--border-question);*/
+    background-color: var(--question-header-bg);
+    border-radius: 6px;
+}
+
 .nbae-question.right, .nbae-quiz .summary.ok {
-    background-color: #d4f8e8;
+    background-color: var(--question-bg-right);
 }
 .nbae-question.wrong, .nbae-quiz .summary.ko {
-    background-color: #ffd6d9;
+    background-color: var(--question-bg-wrong);
 }
 .nbae-question.unanswered {
-    background-color: #f0f0f0;
+    background-color: var(--question-bg-odd);
 }
 .nbae-question.unanswered:nth-child(2n) {
-    background-color: #e8e8e8;
+    background-color: var(--question-bg-even);
 }
 
 .nbae-quiz .submit {
     margin: 10px;
     border-radius: 10px;
-    background-color: #c2f0fc;
-    border: 2px solid #084177;
+    background-color: var(--submit-bg);
+    border: var(--border-submit);
     width: max-content;
 }
 
@@ -95,7 +128,7 @@ CSS = """
 }
 
 .nbae-quiz .wrong-answer {
-    background-color: #ff2e63;
+    background-color: var(--option-wrong);
     border-radius: 6px;
 }
 
@@ -111,10 +144,89 @@ CSS = """
 .nbae-question.wrong span.wrong {
     font-weight: bold;
     font-size: larger;    
+    padding: 4px;
+}
+.nbae-question.unanswered span.unanswered {
+    background-color: var(--question-bg-odd);
+}
+.nbae-question.unanswered:nth-child(2n) span.unanswered {
+    background-color: var(--question-bg-even);
+}
+.nbae-question.right span.right {
+    background-color: var(--question-bg-right);
+}
+.nbae-question.wrong span.wrong {
+    background-color: var(--question-bg-wrong);
+}
+ 
+.nbae-question div.options>div.wrong-answer {
+    margin-top: 8px;
+}
+.nbae-question div.options.widget-vbox>div {
+    border-bottom: 0.8px solid #d0d0d0;
+}
+.nbae-question div.options.widget-hbox>div {
+    border-right: 0.8px solid #d0d0d0;
+}
+
+.nbae-question .explanation::before {
+    content: "☛";
+    margin-right: 10px;
+    font-size: 150%;
+}
+.nbae-question .explanation {
+    border-radius: 6px;
+    padding: 10px;
+    background-color: var(--explanation-right);
+    display: none;
+    border: var(--border-explanation);
+}
+
+.nbae-question .wrong-answer .explanation {
+    background-color: var(--explanation-wrong);
+}
+
+.nbae-question:not(.revealed) .explanation {
+    display: none;
+}
+.nbae-question.revealed .explanation {
+    display: inherit;
+}
+.nbae-question /*div.widget-html-content*/ pre {
+    line-height: 1.2;
+}
+.nbae-question div.widget-html-content {
+    line-height: 1.6;
+}
+.nbae-question div.option-box div {
+    align-self: center;
 }
 """
 
 
+# a flexible content can be defined either with a plain str
+# or with a Content object
+FlexibleContent = Union[str, Content]
+
+class Flexible:
+    # it's important for the YAML loader that this parameter be called text
+    def __init__(self, text: FlexibleContent):
+        if isinstance(text, Content):
+            self.content = text
+        elif isinstance(text, str):
+            self.content = MarkdownContent(text)
+        else:
+            raise ValueError(f"unexpected type {type(text).__name__}"
+                             f" for flexible {text}")
+    def widget(self):
+        return self.content.widget()
+
+
+class Explanation(Flexible):
+    def widget(self):
+        w = super().widget()
+        return w.add_class('explanation') if w else None
+
 
 class GenericBooleanOption:
     """
@@ -138,11 +250,14 @@ class Option(GenericBooleanOption):
     that students should check as such
     
     """
-    def __init__(self, text, **kwds):
+    def __init__(self, text, explanation=None, **kwds):
         super().__init__(**kwds)
         self.text = text
+        self.explanation = explanation
     def render(self):
         return TextContent(self.text)
+    def explanation_widget(self):
+        return None if not self.explanation else self.explanation.widget()
     
 
 class CodeOption(Option):
@@ -171,7 +286,7 @@ class _TeacherOptions:
 
 # this class captures the order in which options or questions 
 # actually displayed, which is randomized from the input list
-# (as defined in the Python code) when shuffle is True
+# (as defined in the YAML code) when shuffle is True
 class _DisplayedItems:
     def __init__(self, input_list, shuffle):
         self.displayed = input_list[:]
@@ -189,14 +304,7 @@ class _DisplayedOptions(_DisplayedItems):
 class _DisplayedQuestions(_DisplayedItems):
     pass
 
-# one can define a question from a plain str or a Content object
-QuestionType = Union[str, Content]
-def question_to_widget(question: QuestionType):
-    if isinstance(question, Content):
-        return question.widget()
-    else:
-        return HTMLMath(question)
-    
+
 
 class Answer(Enum):
     UNANSWERED = -1
@@ -216,16 +324,19 @@ class Score:
         self.if_right = 1
         self.if_wrong = -1
         self.if_unanswered = 0
+        # accept lists for the yaml loader
+        ok_types = (tuple, list)
         if single_arg is None:
             pass
         elif isinstance(single_arg, int):
             self.if_right = single_arg
-        elif isinstance(single_arg, tuple) and len(single_arg) == 2:
+        elif isinstance(single_arg, ok_types) and len(single_arg) == 2:
             self.if_right, self.if_wrong = single_arg
-        elif isinstance(single_arg, tuple) and len(single_arg) == 3:
+        elif isinstance(single_arg, ok_types) and len(single_arg) == 3:
             self.if_right, self.if_wrong, self.if_unanswered = single_arg
         else:
-            raise ValueError("Score constructor expects an int or a tuple of 2/3 ints")
+            raise ValueError(f"Score constructor expects a single int\n"
+                             f"or a {ok_types} of 2 or 3 ints")
 
 
     def score(self, answer):
@@ -266,11 +377,15 @@ class QuizQuestion:
     instead of a vertical one    
     """
     
+    # constructor is kept as slight as possible for the yaml loader
+    # in that context options is not yet a list of Options objects
+    # but just plain Python dicts as outcome from yaml
     def __init__(self, *,
-                 question: QuestionType,
+                 question: FlexibleContent,
                  options: List, 
                  # if defined, show up on top of the alternatives
-                 question2: str=None,
+                 question_sequel: FlexibleContent=None,
+                 explanation=None, 
                  # do we want to shuffle the options
                  shuffle=True, 
                  # set this to True to mak it plain 
@@ -283,24 +398,36 @@ class QuizQuestion:
                  horizontal_layout=False,
                  horizontal_options=False):
         self.question = question
-        self.teacher_options = _TeacherOptions(options)
-        self.question2 = question2
+        self.options = options
+        self.question_sequel = question_sequel
+        self.explanation = explanation
+        self.shuffle = shuffle
         self.exactly_one_option = exactly_one_option
         self._score_object = Score(score)
         self.horizontal_layout = horizontal_layout
         self.horizontal_options = horizontal_options
-        # shuffle if requested
-        self._displayed_options = _DisplayedOptions(options, shuffle)
-        # add 'none of the above' option as last option if requested
-        if option_none is not None:
-            self.teacher_options.append(option_none)
-            self._displayed_options.append(option_none)
+        self.option_none = option_none
         #
         self.feedback_area = None
         self._widget_instance = None
         # the rank in the Quiz object
         self.index = None
+        # 
+        self._post_inited = False
+        
+
+    # post-processing once options has been elaborated
+    def post_init(self):
+        if self._post_inited:
+            return
+        # shuffle if requested
+        self._displayed_options = _DisplayedOptions(self.options, self.shuffle)
+        # add 'none of the above' option as last option if requested
+        if self.option_none is not None:
+            self.options.append(self.option_none)
+            self._displayed_options.append(self.option_none)
         self.sanity_check()
+        self._post_inited = True
         
 
     def sanity_check(self):
@@ -338,42 +465,63 @@ class QuizQuestion:
 
     def widget(self):
         
+        self.post_init()
+        
         if self._widget_instance:
             return self._widget_instance
         
+        # header area
         header_widget = HBox([
             HTML(f'Question # {self.index}').add_class('index'),
             HTML(f'{self._score_object.html()}').add_class('score')
-        ]).add_class('header')
-        question_widget = question_to_widget(self.question)
-        question = VBox([header_widget,
-                         question_widget]).add_class('question')
+        ]).add_class('question-header')
+        question_widget = Flexible(self.question).widget()
+        header_widgets = [header_widget, question_widget]
+        if self.explanation:
+            header_widgets.append(self.explanation.widget()) 
+        question = VBox(header_widgets).add_class('question')
         if self.exactly_one_option:
             question.add_class('exactly-one')
-            
+        
+        # the options per se    
         # it's important that we have as many checkboxes as option_boxes
-        self.checkboxes = [Checkbox(value=option.selected, disabled=False, description='', indent=False)
+        self.checkboxes = [Checkbox(value=option.selected, disabled=False,
+                                    description='', indent=False)
                            for option in self._displayed_options]
+        # radio-box behaviour
         if self.exactly_one_option:
             for checkbox in self.checkboxes:
                 checkbox.observe(lambda event: self.radio_button_callback(event))
         labels = [option.render().widget() for option in self._displayed_options]
+        
+        # explanations contains, for each option, either a widget, or None
+        explanations = [option.explanation_widget() 
+                             for option in self._displayed_options]
+        
         options_box = HBox if self.horizontal_options else VBox
-        self.option_boxes = [HBox([checkbox, label]) 
-                             for (checkbox, label) in zip(self.checkboxes, labels)]
-        if not self.question2:
+        def make_option(checkbox, label, explanation):
+            main = HBox([checkbox, label]).add_class('option-box')
+            return (main if not explanation
+                    else VBox([main, explanation]))
+        self.option_boxes = [
+            make_option(checkbox, label, explanation)
+            for (checkbox, label, explanation) 
+            in zip(self.checkboxes, labels, explanations)]
+        if not self.question_sequel:
             actual_sons = self.option_boxes
         else:
-            actual_sons = [question_to_widget(self.question2).add_class('question2')]
+            actual_sons = [Flexible(self.question_sequel).widget()
+                           .add_class('question-sequel')]
             actual_sons += self.option_boxes
-        answers = options_box(actual_sons)
-        answers.add_class('answers')
+        options = options_box(actual_sons)
+        options.add_class('options')
 
         css_widget = CssContent(CSS).widget()
         
+        # putting it all together
         layout_box = HBox if self.horizontal_layout else VBox
         self._widget_instance = layout_box(
-            [question, answers, css_widget])
+            [question, options, css_widget])
         self._widget_instance.add_class('nbae-question')
         self.feedback_area = self._widget_instance
         self.feedback(Answer.UNANSWERED)
@@ -417,23 +565,24 @@ class QuizQuestion:
 
 
     def individual_feedback(self):
-        for option, checkbox, option_box in zip(
-            self._displayed_options, self.checkboxes, self.option_boxes):
+        for option, option_box, checkbox in zip(
+            self._displayed_options, self.option_boxes, self.checkboxes):
             checkbox.disabled = True            
             # good answer ?
             if option.correct == checkbox.value:
                 option_box.remove_class('wrong-answer')
             else:
                 option_box.add_class('wrong-answer')
+        self._widget_instance.add_class('revealed')
 
 
     def preserve(self) -> List[bool]:
         for option, checkbox in zip(self._displayed_options, self.checkboxes):
             option.selected = checkbox.value
-        return [option.selected for option in self.teacher_options]
+        return [option.selected for option in self.options]
             
     def restore(self, bools: List[bool]):
-        for option, boolean in zip(self.teacher_options, bools):
+        for option, boolean in zip(self.options, bools):
             option.selected = boolean
         if self._widget_instance:
             for checkbox, option in zip(self.checkboxes, self._displayed_options):
@@ -446,35 +595,53 @@ class Quiz:
     one can only submit a full Quiz, not just one question at a time
     """    
     
+    # same approach to hyper-light constructor
+    # the questions attribute might temporarily be a list of
+    # plain Python objects, not QuizQuestion instances yet
     def __init__(self,
-                 exoname, 
                  *,
+                 exoname, 
                  questions: List[QuizQuestion], 
                  shuffle=True,
                  max_attempts=2,
                  max_grade=None):
         self.exoname = exoname
-        self.quiz_questions = questions
+        self.questions = questions
+        self.shuffle = shuffle
         self.max_attempts = max_attempts
         self.max_grade = max_grade
         
-        self.displayed_questions = _DisplayedQuestions(questions, shuffle)
+        # private - for updating the UI
+        self.submit_button = None
+        self.submit_summary = None
+        #
+        self._post_inited = False        
+
+
+    def post_init(self):
+        if self._post_inited:
+            return
+        # this is tricky and not quite right
+        # we need to have the questions post-init'ed
+        # this early because of the restoration below
+        # otherwise questions won't have their option_none appended
+        # and restore would fail to restore the option_none option
+        for question in self.questions:
+            question.post_init()
+        self.displayed_questions = _DisplayedQuestions(self.questions, self.shuffle)
         # needs to be saved somewhere
         self.current_attempts = storage_read(self.exoname, 'current_attempts', 0)
         preserved = storage_read(self.exoname, "answers", [])
         if preserved: 
             self.restore(preserved)
-            
-        # private - for updating the UI
-        self.submit_button = None
-        self.submit_summary = None
-        
         # set question rank
         for index, question in enumerate(self.displayed_questions, 1):
             question.set_index(index)
+        self._post_inited = True
 
 
     def widget(self):
+        self.post_init()
         sons = [question.widget() for question in self.displayed_questions]
 
         self.submit_button = Button(description='submit').add_class('submit')
@@ -489,21 +656,30 @@ class Quiz:
     
     def submit(self, _button):
         self.current_attempts += 1
+        # because of possible load/network latency,
+        #  we need to disable this until the event is processed
+        # update() will re-enable it later on if needed
+        self.submit_button.disabled = True
         self.update()
         storage_save(self.exoname, 'current_attempts', self.current_attempts)
         storage_save(self.exoname, "answers", self.preserve())
         (current_score, max_score,
          normalized_score, normalized_max_score) = self.total_score()
-        log_quiz(self.exoname, current_score, max_score,
-                 normalized_score, normalized_max_score)
-        
-        
-        
+        log_kwds = {}
+        if self.max_grade:
+            log_kwds = dict(normalized_score=normalized_score,
+                            normalized_max_score=normalized_max_score)
+        log2_quiz(self.exoname, 
+                  attempt=self.current_attempts, max_attempts=self.max_attempts,
+                  score=current_score, max_score=max_score,
+                  **log_kwds)
+
+
     def preserve(self) -> List[List[bool]]:
-        return [question.preserve() for question in self.quiz_questions]
+        return [question.preserve() for question in self.questions]
             
     def restore(self, list_of_list_of_bools):
-        for question, list_of_bools in zip(self.quiz_questions, list_of_list_of_bools):
+        for question, list_of_bools in zip(self.questions, list_of_list_of_bools):
             question.restore(list_of_bools)
 
 
@@ -519,6 +695,7 @@ class Quiz:
         
 
     def update(self):
+        # there may be latency if the host is loaded
         self.answers = [question.answer() 
                         for question in self.displayed_questions]
         right_answers = [answer for answer in self.answers if answer == Answer.RIGHT]
@@ -529,7 +706,6 @@ class Quiz:
                 question.feedback(question.answer())
                 question.individual_feedback()
             # disable submit button
-            self.submit_button.disabled = True
             self.submit_button.description = "quiz over"
             summary = self.final_score_html()
             if self.max_attempts > 1:
@@ -540,6 +716,9 @@ class Quiz:
                 self.submit_summary.add_class('ok')
             else:
                 self.submit_summary.add_class('ko')
+            # not really needed (it's been disable already in submit)
+            # but for consistency
+            self.submit_button.disabled = True
         else:
             # so here we know that self.current_attempts < self.max_attempts:
             left = self.max_attempts - self.current_attempts
@@ -548,13 +727,14 @@ class Quiz:
             if self.current_attempts >= 1:
                 self.submit_summary.value = (
                     f"{len(right_answers)}/{len(self.answers)} questions OK")
+            self.submit_button.disabled = False
 
     def total_score(self):
         """
         returns a tuple current_score, max_score
         """
-        current_score = sum(q.score() for q in self.quiz_questions)
-        max_score = sum(q.max_score() for q in self.quiz_questions)
+        current_score = sum(q.score() for q in self.questions)
+        max_score = sum(q.max_score() for q in self.questions)
         if self.max_grade is not None:
             normalized_max_score = self.max_grade
             normalized_score = current_score/max_score * normalized_max_score
